# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper classes and functions that can be used with more than one graph type.
"""

import collections
import enum
import functools
import logging
import time
import warnings

import networkx

import graphcat.optional
import graphcat.require


numpy = graphcat.optional.module("numpy")


log = logging.getLogger(__name__)


class Array(object):
    """Task function callable that returns a caller-supplied array.

    Parameters
    ----------
    value: :class:`numpy.ndarray` convertable object, required
        The array to use as the output for this callable.

    See Also
    --------
    :func:`array` - factory function for :class:`Array` instances.
    """
    def __init__(self, value):
        self._value = value

    def __call__(self, graph, name, inputs, extent=None):
        return self._value[extent] if extent is not None else self._value

    def __eq__(self, other):
        return type(self) is type(other) and numpy.all(self._value == other._value)


class ArrayExtent(object):
    """Helper for creating :class:`Array` compatible streaming extents.

    To generate extents, use any numpy-compatible
    `indexing notation <https://numpy.org/doc/stable/reference/arrays.indexing.html>`_::

        extent = ArrayExtent[0:4096]
        extent = ArrayExtent[::2]
        extent = ArrayExtent[:, 0]
        ...

    These extents are compatible with :class:`Array`.
    """
    def __class_getitem__(cls, key):
        return key


class Constant(object):
    """Task function callable that returns a caller-supplied value.

    Parameters
    ----------
    value: any Python object, required
        The value to use as the output for this callable.

    See Also
    --------
    :func:`constant` - factory function for :class:`Constant` instances.
    """
    def __init__(self, value):
        self._value = value

    def __call__(self, graph, name, inputs, extent=None):
        return self._value

    def __eq__(self, other):
        return type(self) is type(other) and self._value == other._value


class Delay(object):
    """Task function callable that sleeps for a fixed time.

    This is mainly useful for testing and debugging.

    Parameters
    ----------
    seconds: number, required
        The number of seconds to sleep when executed.

    See Also
    --------
    :func:`delay` - factory function for :class:`Delay` instances.
    """
    def __init__(self, seconds):
        self._seconds = seconds

    def __call__(self, graph, name, inputs, extent=None):
        time.sleep(self._seconds)

    def __eq__(self, other):
        return type(self) is type(other) and self._seconds == other._seconds


class DeprecationWarning(Warning):
    """Warning category for deprecated code."""
    pass


class Input(enum.Enum):
    """Enumerates special :class:`graphcat.graph.Graph` named inputs."""
    IMPLICIT = 1
    """Named input for links that are generated automatically for use as implicit dependencies, not data sources."""


class Logger(object):
    """Log graph events.

    Create a :class:`Logger` object to see the behavior of
    the graph during updates, using the Python :mod:`logging`
    module::

        logger = graphcat.Logger(graph)

    This is useful for debugging and pedagogy.  The logger will
    generate output for five types of event:

    * cycle - a cycle is detected during updating.
    * executed - a task is executed.
    * failed - a task raises an exception during execution.
    * finished - a task executes successfully.
    * updated - a task is updated.

    Update events happen regardless of the state of a task.  Execute events
    only happen if the task isn't already finished.  Failed and finished events
    only happen if a task is executed.

    Parameters
    ----------
    graph: class:`graphcat.graph.Graph`, required
        The graph whose events will be logged.
    """
    def __init__(self, graph, log_exceptions=True, log_inputs=True, log_outputs=True, log_extents=True, log=log):
        self._log_exceptions = log_exceptions
        self._log_inputs = log_inputs
        self._log_outputs = log_outputs
        self._log_extents = log_extents
        self._log = log

        graph.on_cycle.connect(self.on_cycle)
        graph.on_execute.connect(self.on_execute)
        graph.on_failed.connect(self.on_failed)
        graph.on_finished.connect(self.on_finished)
        graph.on_update.connect(self.on_update)


    def on_cycle(self, graph, name):
        """Called when a cycle is detected."""
        self._log.info(f"Task {name} cycle detected.")

    def on_execute(self, graph, name, inputs, extent=None):
        """Called when a task is executed."""
        message = f"Task {name} executing."
        if self._log_inputs:
            message += f" Inputs: {inputs}"
        if self._log_extents and graph.is_streaming:
            message += f" Extent: {extent}"
        self._log.info(message)

    def on_failed(self, graph, name, exception):
        """Called when a task raises an exception during execution."""
        if self._log_exceptions:
            self._log.error(f"Task {name} failed. Exception: {exception}")
        else:
            self._log.error(f"Task {name} failed.")

    def on_finished(self, graph, name, output):
        """Called when a task has executed sucessfully."""
        if self._log_outputs:
            self._log.info(f"Task {name} finished. Output: {output}")
        else:
            self._log.info(f"Task {name} finished.")

    def on_update(self, graph, name):
        """Called when a task is updated."""
        message = f"Task {name} updating."
        self._log.info(message)


class Passthrough(object):
    """Task function callable that always returns an upstream input.

    Parameters
    ----------
    input: hashable object, required
        Name of the input to be returned when this task is executed.  Note that
        there must be exactly one connection to the named input, or an
        exception will be raised.

    See Also
    --------
    :func:`passthrough` - factory function for :class:`Passthrough` instances.
    """
    def __init__(self, input):
        self._input = input

    def __call__(self, graph, name, inputs, extent=None):
        return inputs.getone(self._input)

    def __eq__(self, other):
        return type(self) is type(other) and self._input == other._input


class PerformanceMonitor(object):
    """Tracks the performance of graph tasks as they're executed.

    Parameters
    ----------
    graph: :class:`graphcat.graph.Graph`, required
        Graph whose performance will be monitored.
    """
    def __init__(self, graph):
        self.reset()
        graph.on_execute.connect(self._on_execute)
        graph.on_failed.connect(self._on_failed)
        graph.on_finished.connect(self._on_finished)


    def _on_execute(self, graph, name, inputs, extent=None):
        self._start = time.time()


    def _on_failed(self, graph, name, exception):
        self._tasks[name].append(time.time() - self._start) # pragma: no cover


    def _on_finished(self, graph, name, output):
        self._tasks[name].append(time.time() - self._start)


    def reset(self):
        """Clear performance data."""
        self._tasks = collections.defaultdict(list)


    @property
    def tasks(self):
        """Graph task execution times since this object was created / reset.

        Returns
        -------
        tasks: :class:`dict` containing :class:`list` values.
            Maps the name of every task that has been updated to an array
            containing execution times.
        """
        return dict(self._tasks)


class RaiseException(object):
    """Task function callable that raises an exception when executed.

    This is mainly useful for testing and debugging.

    Parameters
    ----------
    exception: :class:`Exception`, required
        The exception to be raised when the task is executed.

    See Also
    --------
    :func:`raise_exception` - factory function for :class:`RaiseException` instances.
    """
    def __init__(self, exception):
        self._exception = exception

    def __call__(self, graph, name, inputs, extent=None):
        raise self._exception

    def __eq__(self, other):
        return type(self) is type(other) and self._exception == other._exception


class TaskState(enum.Enum):
    """Enumerates :class:`graphcat.graph.Graph` task states.

    Every task within a :class:`graphcat.graph.Graph` will always be in one of
    the following states.

    See Also
    --------
    :meth:`graphcat.graph.Graph.state` - returns the state of a task.
    """
    UNFINISHED = 1
    """The task is out-of-date and will be executed during the next update."""
    FAILED = 2
    """The task or one of it's dependencies failed during the last update."""
    FINISHED = 3
    """The task executed successfully during the last update."""


class UpdatedTasks(object):
    """Maintains a list of graph tasks that have been updated.

    Parameters
    ----------
    graph: :class:`graphcat.graph.Graph`, required
        Graph to watch for task updates.
    """
    def __init__(self, graph):
        self._tasks = set()
        graph.on_update.connect(self._on_update)

    def _on_update(self, graph, name):
        self._tasks.add(name)

    @property
    def tasks(self):
        """Graph tasks that have received updates since this object was created.

        Returns
        -------
        tasks: :class:`set`
            Python :class:`set` containing the name of every task that has been
            updated.
        """
        return self._tasks


graphcat.require.loaded_module("numpy")
def array(value):
    """Factory for task functions that return array values when executed.

    Note
    ----
    This callable is designed to be compatible with :class:`ArrayExtent` extents
    when used in a :class:`graphcat.streaming.StreamingGraph`.

    Parameters
    ----------
    value: :class:`numpy.ndarray`-convertable value, required
        The array to return when the task is executed.

    Returns
    -------
    fn: :class:`Array`
        Task function that will always return `value` when executed.
    """
    return Array(value)


def automatic_dependencies(fn):
    """Function decorator that automatically tracks dependencies.

    Use this to decorate task functions that need dependency tracking, such as
    :func:`evaluate`.

    See Also
    --------
    :meth:`graphcat.graph.Graph.set_expression`
        Convenience method that configures a task to evaluate expressions and
        automatically track dependencies.
    """
    @functools.wraps(fn)
    def implementation(graph, name, inputs, extent=None):
        # Remove old implicit dependencies.
        edges = list(graph._graph.out_edges(name, data="input", keys=True))
        for target, source, key, input in edges:
            if input == Input.IMPLICIT:
                graph._graph.remove_edge(target, source, key)

        # Keep track of all dependencies while the task executes.
        updated = UpdatedTasks(graph)
        result = fn(graph, name, inputs, extent)

        # Filter out dependencies that are already explicitly captured.
        dependencies = updated.tasks
        dependencies = dependencies.difference(networkx.descendants(graph._graph, name))
        dependencies = dependencies.difference([name])

        # Create new implicit dependencies with what remains.
        for source in dependencies:
            graph._graph.add_edge(name, source, input=Input.IMPLICIT)

        return result
    return implementation


def builtins(graph, name, inputs, extent=None):
    """Returns standard builtin symbols for expression tasks.

    See Also
    --------
    :func:`evaluate` - uses :func:`builtins` as the default for the `symbols` argument.
    """
    return {
        "graph": graph,
        "name": name,
        "inputs": inputs,
        "extent": extent,
    }


def constant(value):
    """Factory for task functions that return constant values when executed.

    This is useful when creating a task that will act as a parameter
    for a downstream task::

        graph.add_task("theta", constant(math.pi))

    To change the parameter later, use :func:`constant` again, with
    :meth:`Graph.set_task_fn` to specify a new function::

        graph.set_task_fn("theta", constant(math.pi / 2))

    Parameters
    ----------
    value: any value, required
        The value to return when the task is executed.

    Returns
    -------
    fn: :class:`Constant`
        Task function that will always return `value` when executed.
    """
    return Constant(value)


def consume(graph, name, inputs, extent=None):
    """Task function that retrieves all its inputs, but otherwise does nothing.

    This is mainly useful for debugging :class:`dynamic graphs<graphcat.dynamic.DynamicGraph>`,
    since the default :func:`null` task function won't execute upstream nodes.
    """
    values = [value() for value in inputs.values()]


def delay(seconds):
    """Factory for task functions that sleep for a fixed time.

    This is mainly useful for testing and debugging.

    Parameters
    ----------
    seconds: number, required
        Number of seconds to sleep when executed.

    Returns
    -------
    fn: function
        Task function that will always sleep for `seconds` when executed.
    """
    return Delay(seconds)


def evaluate(code, symbols=None):
    """Factory for task functions that evaluate Python expressions.

    If your expressions can access the output from other tasks in the graph,
    you will want to use this function with the :func:`automatic_dependencies`
    decorator, or use :meth:`graphcat.graph.Graph.set_expression` which sets up dependency
    tracking for you.

    See Also
    --------
    :meth:`graphcat.graph.Graph.set_expression`
        Convenience method that configures a task to evaluate expressions and
        automatically track dependencies.

    Parameters
    ----------
    code: string, required
        Python code to be executed when the task is executed.
    symbols: callable, optional
        Function that returns a Python dict containing symbols that will be
        available to the expression when it's executed.  If :any:`None` (the
        default), the :func:`builtins` function will be used, which gives the
        expression access to the same `graph`, `name`, `inputs`, and `extent`
        objects as a normal task function.

    Returns
    -------
    fn: function
        Task function that will execute Python code when the
        task is executed.
    """
    if symbols is None:
        symbols = builtins

    def implementation(graph, name, inputs, extent=None):
        try:
            return eval(code, {}, symbols(graph, name, inputs, extent))
        except Exception as e: # pragma: no cover
            raise RuntimeError(f"Uncaught exception executing expression {code!r}: {e}")
    return implementation


def null(graph, name, inputs, extent=None):
    """Task function that does nothing.

    This is the default if you don't specify a function for
    :meth:`Graph.add_task` or :meth:`Graph.set_task_fn`, and is useful in
    debugging and pedagogy.
    """
    pass


def passthrough(input=None):
    """Factory for task functions that pass-through incoming data.

    Callers can use this function to temporarily bypass tasks in the graph.

    Parameters
    ----------
    input: hashable object, required
        The named input that will pass-through to the task output.

    Returns
    -------
    fn: function
        Task function that will pass the input value named `input` to its output.
    """
    return Passthrough(input)


def raise_exception(exception):
    """Factory for task functions that raise an exception when executed.

    This is mainly useful for testing and debugging.

    Parameters
    ----------
    exception: :class:`BaseException` derivative, required
        The exception to raise when the task is executed.

    Returns
    -------
    fn: function
        Task function that will always raise `exception` when executed.
    """
    return RaiseException(exception)




# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from behave import *

import os
import pkgutil
import subprocess
import sys

root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
docs_dir = os.path.join(root_dir, "docs")

copyright_notice = """# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""


@given(u'all sources.')
def step_impl(context):
    context.sources = []
    for directory, subdirectories, filenames in os.walk(root_dir):
        for filename in filenames:
            if os.path.splitext(filename)[1] not in [".py"]:
                continue

            context.sources.append(os.path.join(directory, filename))
    context.sources = sorted(context.sources)


@then(u'every source must contain a copyright notice.')
def step_impl(context):
    for source in context.sources:
        with open(source, "r") as fobj:
            if not fobj.read().startswith(copyright_notice):
                raise AssertionError("%s missing copyright notice." % source)





# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements computational graphs using dynamic dependency analysis.
"""

import functools

import graphcat.common
import graphcat.graph


class DynamicGraph(graphcat.graph.Graph):
    """Manages a dynamic computational graph.

    The graph is a collection of named tasks, connected by links that define
    dependencies between tasks.  Updating a task implicitly updates all of its
    transitive dependencies.  When an unfinished task is updated, it executes a
    user-supplied function and stores the function return value as the task
    output.  Outputs of upstream tasks are automatically passed as inputs to
    downstream tasks.
    """
    def __init__(self):
        super().__init__()


    def _add_node(self, name, fn):
        self._graph.add_node(name, fn=fn, state=graphcat.common.TaskState.UNFINISHED, output=None, updating=False)


    def _mark_unfinished(self, name):
        node = self._graph.nodes[name]
        node["output"] = None
        node["state"] = graphcat.common.TaskState.UNFINISHED


    def _output(self, name):
        self._update(name)
        return self._graph.nodes[name]["output"]


    def _update(self, name):
        # Break cycles
        task = self._graph.nodes[name]
        if task["updating"]:
            self._on_cycle.send(self, name=name)
            return
        task["updating"] = True

        # Notify observers that the task will be updated.
        self._on_update.send(self, name=name)

        # Only execute this task if it isn't already finished.
        if task["state"] != graphcat.common.TaskState.FINISHED:
            try:
                # Get the task inputs.
                inputs = NamedInputs(self, name)

                # Execute the function and store the output.
                self._on_execute.send(self, name=name, inputs=inputs)
                task["output"] = task["fn"](graph=self, name=name, inputs=inputs)
                task["state"] = graphcat.common.TaskState.FINISHED
                self._on_finished.send(self, name=name, output=task["output"])
            except Exception as e:
                # The function raised an exception, notify observers.
                task["output"] = None
                task["state"] = graphcat.common.TaskState.FAILED
                self._on_failed.send(self, name=name, exception=e)
                task["updating"] = False
                raise e

        task["updating"] = False


    @property
    def is_dynamic(self):
        """Returns :any:`True`."""
        return True


    @property
    def is_streaming(self):
        """Returns :any:`False`."""
        return False


    def output(self, name):
        """Retrieve the output from a task.

        This implicitly updates the graph, so the returned value is
        guaranteed to be up-to-date.

        Parameters
        ----------
        name: hashable object, required
            Unique task name.

        Returns
        -------
        output: any object
            The value returned when the task function was last executed, or :any:`None`.

        Raises
        ------
        :class:`ValueError`
            If `name` doesn't exist.
        :class:`Exception`
            Any exception raised by a task function will be re-raised by :meth:`output`.
        """
        self._require_task_present(name)
        self.update(name)
        return self._graph.nodes[name]["output"]


    def update(self, name):
        """Update a task and all of its transitive dependencies.

        Parameters
        ----------
        name: hashable object, required
            Name identifying the task to be updated.

        Raises
        ------
        :class:`ValueError`
            If the task with `name` doesn't exist.
        :class:`Exception`
            Any exception raised by a task function will be re-raised by :meth:`update`.
        """

        self._require_task_present(name)
        self._update(name)


class NamedInputs(object):
    """Access named inputs for a graph task.

    Parameters
    ----------
    graph: :class:`DynamicGraph`, required
        Graph containing a task.
    name: hashable object, required
        Existing task unique name.
    """
    def __init__(self, graph, name):
        if not isinstance(graph, DynamicGraph):
            raise ValueError("Graph input must be an instance of DynamicGraph") # pragma: no cover

        edges = graph._graph.out_edges(name, data="input")
        self._keys = [input for target, source, input in edges]
        self._values = [functools.partial(graph._output, source) for target, source, input in edges]

    def __contains__(self, name):
        """Return :any:`True` if `name` matches a named input for this task."""
        return name in self._keys

    def __len__(self):
        """Return the number of named inputs for this task."""
        return len(self._keys)

    def __repr__(self):
        inputs = ", ".join([repr(key) for key in self._keys])
        return f"{{{inputs}}}"

    def get(self, name, default=None):
        """Return a single input value.

        Use this method to return a value when you expect to have either zero
        or one input that matches `name`.

        Parameters
        ----------
        name: hashable object, required
            Name of the input value to return.
        default: any Python value, optional
            If an input matching `name` doesn't exist, this value will be
            returned instead.  Defaults to :any:`None`.

        Returns
        -------
        value: any Python value
            The value of input `name`, or `default`.

        Raises
        ------
        :class:`KeyError`: if more than one input matches `name`.
        """
        values = [value for key, value in zip(self._keys, self._values) if key == name]
        if len(values) == 0:
            return default
        elif len(values) == 1:
            return values[0]()
        else:
            raise KeyError(f"More than one input {name!r}")

    def getall(self, name):
        """Return multiple input values.

        Use this method to return every input value that matches `name`.

        Parameters
        ----------
        name: hashable object, required
            Name of the input value to return.

        Returns
        -------
        values: list of Python values
            Values from every input that matches `name`.  Returns an empty list
            if there are none.
        """
        return [value() for key, value in zip(self._keys, self._values) if key == name]

    def getone(self, name):
        """Return a single input value.

        Use this method to return a value when you expect to have exactly one
        input that matches `name`.

        Parameters
        ----------
        name: hashable object, required
            Name of the input value to return.

        Returns
        -------
        value: any Python value
            The value of input `name`.

        Raises
        ------
        :class:`KeyError`: if more or less than one input matches `name`.
        """
        values = [value for key, value in zip(self._keys, self._values) if key == name]
        if len(values) == 0:
            raise KeyError(name)
        elif len(values) == 1:
            return values[0]()
        else:
            raise KeyError(f"More than one input {name!r}")

    def items(self):
        """Return names and values for every input attached to this task.

        Note
        ----
        For each (name, value) pair returned by this method, the value is a
        callable that returns the actual value from the upstream task.

        Returns
        -------
        values: sequence of (hashable object, callable) tuples
            The name and value of every input attached to this task.
        """
        return zip(self._keys, self._values)

    def keys(self):
        """Return names for every input attached to this task.

        Returns
        -------
        names: sequence of hashable objects
            The name of every input attached to this task.  Note that the same
            name may appear more than once in the sequence.
        """
        return self._keys

    def values(self):
        """Return values for every input attached to this task.

        Note
        ----
        Each value returned by this method is a callable that returns the
        actual value from the upstream task.

        Returns
        -------
        values: sequence of callables
            The value of every input attached to this task, in the same
            order as :meth:`keys`.
        """
        return self._values





# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for implementing optional functionality.
"""


def module(name):
    """Quietly load a module by name, ignoring errors.

    Note that dotted names, e.g. `pkg.mod` will return the
    top-level package, just like the `import` statement.

    Parameters
    ----------
    name: :class:`str`
        Name of the module to be loaded.

    Returns
    -------
    module: loaded module if successful, or :any:`None`.
    """
    try:
        return __import__(name, globals(), locals(), [], 0)
    except: # pragma: no cover
        pass



# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import unittest.mock

from behave import *

import graphcat
import graphcat.diagram
import graphcat.notebook

try:
    import pygraphviz
except:
    pass

try:
    import IPython
except:
    pass

try:
    import numpy
    import numpy.testing
except:
    pass

import test


class EventRecorder(object):
    def __init__(self, graph):
        self.changed = []
        self.cycles = []
        self.exceptions = []
        self.executed = []
        self.failed = []
        self.finished = []
        self.inputs = []
        self.outputs = []
        self.task_renamed = []
        self.updated = []

        graph.on_changed.connect(self.on_changed)
        graph.on_cycle.connect(self.on_cycle)
        graph.on_execute.connect(self.on_execute)
        graph.on_failed.connect(self.on_failed)
        graph.on_finished.connect(self.on_finished)
        graph.on_task_renamed.connect(self.on_task_renamed)
        graph.on_update.connect(self.on_update)

    def on_changed(self, graph):
        self.changed.append(graph)

    def on_cycle(self, graph, name):
        self.cycles.append(name)

    def on_execute(self, graph, name, inputs, extent=None):
        self.executed.append(name)
        self.inputs.append(inputs)

    def on_failed(self, graph, name, exception):
        self.failed.append(name)
        self.exceptions.append(exception)

    def on_finished(self, graph, name, output):
        self.finished.append(name)
        self.outputs.append(output)

    def on_task_renamed(self, graph, oldname, newname):
        self.task_renamed.append((oldname, newname))

    def on_update(self, graph, name, extent=None):
        self.updated.append(name)


#################################################################
# Givens

@given(u'the {module} module is available')
def step_impl(context, module):
    if module not in sys.modules:
        context.scenario.skip()


@given(u'an empty dynamic graph')
def step_impl(context):
    context.graph = graphcat.DynamicGraph()


@given(u'an empty static graph')
def step_impl(context):
    context.graph = graphcat.StaticGraph()


@given(u'an empty streaming graph')
def step_impl(context):
    context.graph = graphcat.StreamingGraph()


@given(u'a log')
def step_impl(context):
    context.log = unittest.mock.Mock()


@given(u'a graph logger')
def step_impl(context):
    context.logger = graphcat.Logger(context.graph, log=context.log)


@given(u'a graph logger with detailed outputs disabled')
def step_impl(context):
    context.logger = graphcat.Logger(context.graph, log_exceptions=False, log_inputs=False, log_outputs=False, log_extents=False, log=context.log)


@given(u'a performance monitor')
def step_impl(context):
    context.performance_monitor = graphcat.PerformanceMonitor(context.graph)


#################################################################
# Whens


@when(u'adding an expression task {name} with expression {expression}')
def step_impl(context, name, expression):
    name = eval(name)
    expression = eval(expression)
    context.events = EventRecorder(context.graph)
    context.graph.set_expression(name, expression)


@when(u'changing the expression task {name} to expression {expression}')
def step_impl(context, name, expression):
    name = eval(name)
    expression = eval(expression)
    context.events = EventRecorder(context.graph)
    context.graph.set_expression(name, expression)


@when(u'adding tasks {names} with functions {functions}')
def step_impl(context, names, functions):
    names = eval(names)
    functions = eval(functions)
    context.events = EventRecorder(context.graph)
    for name, function in zip(names, functions):
        context.graph.add_task(name, function)


@when(u'adding task {name} an exception should be raised')
def step_impl(context, name):
    name = eval(name)
    context.events = EventRecorder(context.graph)
    with test.assert_raises(ValueError):
        context.graph.add_task(name)


@when(u'adding link {link} an exception should be raised')
def step_impl(context, link):
    source, target = eval(link)
    context.events = EventRecorder(context.graph)
    with test.assert_raises(ValueError):
        context.graph.add_links(source, target)


@when(u'adding links {links}')
def step_impl(context, links):
    links = eval(links)
    context.events = EventRecorder(context.graph)
    for source, targets in links:
        context.graph.add_links(source, targets)


@when(u'setting links {links}')
def step_impl(context, links):
    links = eval(links)
    context.events = EventRecorder(context.graph)
    for source, targets in links:
        context.graph.set_links(source, targets)


@when(u'setting parameter {target} {input} {source} {value}')
def step_impl(context, source, value, target, input):
    source = eval(source)
    value = eval(value)
    target = eval(target)
    input = eval(input)
    context.events = EventRecorder(context.graph)
    context.graph.set_parameter(target, input, source, value)


@when(u'adding tasks {names}')
def step_impl(context, names):
    names = eval(names)
    context.events = EventRecorder(context.graph)
    for name in names:
        context.graph.add_task(name)


@when(u'updating task {name} with extent {extent} an exception should be raised')
def step_impl(context, name, extent):
    name = eval(name)
    extent = eval(extent)
    context.events = EventRecorder(context.graph)
    with test.assert_raises(RuntimeError):
        context.graph.update(name, extent=extent)


@when(u'updating task {name} an exception should be raised')
def step_impl(context, name):
    name = eval(name)
    context.events = EventRecorder(context.graph)
    with test.assert_raises(RuntimeError):
        context.graph.update(name)


@when(u'renaming tasks {oldnames} as {newnames}')
def step_impl(context, oldnames, newnames):
    oldnames = eval(oldnames)
    newnames = eval(newnames)
    context.events = EventRecorder(context.graph)
    for oldname, newname in zip(oldnames, newnames):
        context.graph.rename_task(oldname, newname)


@when(u'removing link {link} no exception should be raised')
def step_impl(context, link):
    source, target = eval(link)
    context.events = EventRecorder(context.graph)
    context.graph.clear_links(source, target)


@when(u'removing links {links}')
def step_impl(context, links):
    links = eval(links)
    context.events = EventRecorder(context.graph)
    for source, target in links:
        context.graph.clear_links(source, target)


@when(u'updating tasks {names} with extents {extents}')
def step_impl(context, names, extents):
    names = eval(names)
    extents = eval(extents)
    context.events = EventRecorder(context.graph)
    for name, extent in zip(names, extents):
        context.graph.update(name, extent=extent)


@when(u'updating tasks {names}')
def step_impl(context, names):
    names = eval(names)
    context.events = EventRecorder(context.graph)
    for name in names:
        context.graph.update(name)


@when(u'removing tasks {names} with clear_tasks')
def step_impl(context, names):
    names = eval(names)
    context.events = EventRecorder(context.graph)
    context.graph.clear_tasks(names)


@when(u'the task {name} function is changed to {function}')
def step_impl(context, name, function):
    name = eval(name)
    function = eval(function)
    context.events = EventRecorder(context.graph)
    context.graph.set_task(name, function)


@when(u'tasks {names} are marked unfinished')
def step_impl(context, names):
    names = eval(names)
    for name in names:
        context.graph.mark_unfinished(name)


@when(u'the performance monitor is reset')
def step_impl(context):
    context.performance_monitor.reset()


@when(u'filtering the graph with {hide} then the remaining nodes should match {names}')
def step_impl(context, hide, names):
    hide = eval(hide)
    names = eval(names)

    remaining = [name for name in context.graph.tasks() if not hide(context.graph, name)]
    test.assert_equal(sorted(remaining), sorted(names))


@when(u'computing the task {names} outputs with extents {extents}')
def step_impl(context, names, extents):
    names = eval(names)
    extents = eval(extents)
    context.events = EventRecorder(context.graph)
    context.outputs = [context.graph.output(name, extent=extent) for name, extent in zip(names, extents)]


@when(u'computing the task {names} outputs')
def step_impl(context, names):
    names = eval(names)
    context.events = EventRecorder(context.graph)
    context.outputs = [context.graph.output(name) for name in names]


@when(u'the graph is converted to a diagram')
def step_impl(context):
    context.agraph = graphcat.diagram.draw(context.graph)


#################################################################
# Thens

@then(u'the graph should contain tasks {names}')
def step_impl(context, names):
    names = eval(names)
    test.assert_equal(context.graph.tasks(), set(names))


@then(u'the graph should contain links {links}')
def step_impl(context, links):
    links = eval(links)
    test.assert_equal(sorted(links), sorted(context.graph.links()))


@then(u'the outputs should be {outputs}')
def step_impl(context, outputs):
    outputs = eval(outputs)
    test.assert_equal(outputs, context.outputs)


@then(u'the numpy outputs should be {outputs}')
def step_impl(context, outputs):
    outputs = eval(outputs)
    for a, b in zip(outputs, context.outputs):
        numpy.testing.assert_allclose(a, b)


@then(u'the task {names} state is failed')
def step_impl(context, names):
    names = eval(names)
    for name in names:
        test.assert_equal(context.graph.state(name), graphcat.TaskState.FAILED)


@then(u'the task {names} state is finished')
def step_impl(context, names):
    names = eval(names)
    for name in names:
        test.assert_equal(context.graph.state(name), graphcat.TaskState.FINISHED)


@then(u'the task {names} state is unfinished')
def step_impl(context, names):
    names = eval(names)
    for name in names:
        test.assert_equal(context.graph.state(name), graphcat.TaskState.UNFINISHED)


@then(u'tasks {oldnames} should be renamed to {newnames}')
def step_impl(context, oldnames, newnames):
    oldnames = eval(oldnames)
    newnames = eval(newnames)
    test.assert_equal(list(zip(oldnames, newnames)), context.events.task_renamed)


@then(u'tasks {names} are updated')
def step_impl(context, names):
    names = eval(names)
    test.assert_equal(names, context.events.updated)


@then(u'task {name} has {count} inputs')
def step_impl(context, name, count):
    name = eval(name)
    count = eval(count)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    test.assert_equal(count, len(inputs))


@then(u'the task {name} inputs contain {key}')
def step_impl(context, name, key):
    name = eval(name)
    key = eval(key)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    test.assert_true(key in inputs)


@then(u'the task {name} inputs do not contain {key}')
def step_impl(context, name, key):
    name = eval(name)
    key = eval(key)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    test.assert_true(key not in inputs)


@then(u'getting input {key} from task {name} returns {value}')
def step_impl(context, key, name, value):
    key = eval(key)
    name = eval(name)
    value = eval(value)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    test.assert_equal(inputs.get(key), value)


@then(u'getting input {key} from task {name} raises {exception}')
def step_impl(context, key, name, exception):
    key = eval(key)
    name = eval(name)
    exception = eval(exception)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    with test.assert_raises(exception):
        inputs.get(key)


@then(u'getting one input from task {name} input {key} returns {value}')
def step_impl(context, key, name, value):
    key = eval(key)
    name = eval(name)
    value = eval(value)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    test.assert_equal(inputs.getone(key), value)


@then(u'getting one input from task {name} input {key} raises {exception}')
def step_impl(context, key, name, exception):
    key = eval(key)
    name = eval(name)
    exception = eval(exception)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    with test.assert_raises(exception):
        inputs.getone(key)


@then(u'getting all inputs from task {name} input {key} returns {values}')
def step_impl(context, name, key, values):
    key = eval(key)
    name = eval(name)
    values = eval(values)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    test.assert_equal(values, inputs.getall(key))


@then(u'getting the task {name} input keys returns {values}')
def step_impl(context, name, values):
    name = eval(name)
    values = eval(values)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    test.assert_equal(values, list(inputs.keys()))


@then(u'getting the task {name} input values returns {values}')
def step_impl(context, name, values):
    name = eval(name)
    values = eval(values)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    test.assert_equal(values, [value() for value in inputs.values()])


@then(u'getting the task {name} input items returns {values}')
def step_impl(context, name, values):
    name = eval(name)
    values = eval(values)
    index = context.events.executed.index(name)
    inputs = context.events.inputs[index]
    test.assert_equal(values, [(key, value()) for key, value in inputs.items()])


@then(u'tasks {names} detect cycles')
def step_impl(context, names):
    names = eval(names)
    test.assert_equal(names, context.events.cycles)


@then(u'tasks {names} are executed')
def step_impl(context, names):
    names = eval(names)
    test.assert_equal(names, context.events.executed)


@then(u'tasks {names} are finished')
def step_impl(context, names):
    names = eval(names)
    test.assert_equal(names, context.events.finished)


@then(u'the log should contain {calls}')
def step_impl(context, calls):
    calls = eval(calls)

    calls = [(call[0], (call[1],), {}) for call in calls]
    mock_calls = [tuple(call) for call in context.logger._log.mock_calls]
    test.assert_equal(calls, mock_calls)
    context.logger._log.reset_mock()


@then(u'the graph can be drawn as a diagram')
def step_impl(context):
    agraph = graphcat.diagram.draw(context.graph)


@then(u'displaying the graph in a notebook should produce a visualization')
def step_impl(context):
    graphcat.notebook.display(context.graph)


@then(u'the diagram can be displayed in a notebook')
def step_impl(context):
    graphcat.notebook.display(context.agraph)


@then(u'the graph can be drawn as a diagram with performance overlay')
def step_impl(context):
    agraph = graphcat.diagram.draw(context.graph)
    agraph = graphcat.diagram.performance(agraph, context.performance_monitor)


@then(u'tasks {names} should have links {links}')
def step_impl(context, names, links):
    names = eval(names)
    links = eval(links)
    test.assert_equal(sorted(links), sorted(context.graph.links(names)))


@then(u'testing if the graph contains task {task} should return {result}')
def step_impl(context, task, result):
    task = eval(task)
    result = eval(result)
    test.assert_equal(task in context.graph, result)


@then(u'the performance monitor output should be {outputs}')
def step_impl(context, outputs):
    outputs = eval(outputs)

    monitor = context.performance_monitor
    test.assert_dict_list_values_close(outputs, monitor.tasks, places=None, delta=0.01)





# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

# Repackaging existing tests for PEP-8.

def assert_almost_equal(first, second, places=7, delta=None, msg=None):
    return unittest.TestCase().assertAlmostEqual(first, second, places=places, msg=msg, delta=delta)

def assert_dict_equal(first, second, msg=None):
    return unittest.TestCase().assertDictEqual(first, second, msg)

def assert_equal(first, second, msg=None):
    return unittest.TestCase().assertEqual(first, second, msg)

def assert_is_instance(obj, cls, msg=None):
    return unittest.TestCase().assertIsInstance(obj, cls, msg)

def assert_raises(exception, *, msg=None):
    return unittest.TestCase().assertRaises(exception, msg=msg)

def assert_sequence_equal(first, second, msg=None, seq_type=None):
    return unittest.TestCase().assertSequenceEqual(first, second, msg, seq_type)

def assert_true(expr, msg=None):
    return unittest.TestCase().assertTrue(expr, msg)

# Custom tests.

def assert_dict_keys_equal(first, second, msg=None):
    assert_is_instance(first, dict, "First argument is not a dictionary.")
    assert_is_instance(second, dict, "Second argument is not a dictionary.")
    assert_sequence_equal(first.keys(), second.keys(), "Dict keys do not match.")

def assert_dict_list_values_close(first, second, places=7, delta=None, msg=None):
    assert_dict_keys_equal(first, second, msg)
    for val1, val2 in zip(first.values(), second.values()):
        for el1, el2 in zip(val1, val2):
            assert_almost_equal(el1, el2, places, delta, msg)



# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for testing preconditions and assertions.
"""

import functools
import sys


def loaded_module(modules):
    """Function decorator that tests whether module(s) have already been loaded.

    Parameters
    ----------
    modules: :class:`str` or sequence of :class:`str`, required
        Names of the modules that must already be loaded for the wrapped
        function to execute.

    Raises
    ------
    :class:`RuntimeError`
        If any module in `modules` isn't already loaded.
    """
    if isinstance(modules, str):
        modules = (modules,) # pragma: no cover
    def implementation(f):
        @functools.wraps(f)
        def implementation(*args, **kwargs):
            for module in modules:
                if module not in sys.modules:
                    raise RuntimeError(f"Module {module} could not be found.") # pragma: no cover
            return f(*args, **kwargs)
        return implementation
    return implementation




# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionality for drawing diagrams of computational graphs."""

import graphcat.optional
import graphcat.require


pygraphviz = graphcat.optional.module("pygraphviz")


@graphcat.require.loaded_module("pygraphviz")
def draw(graph, hide=None, rankdir="LR"):
    """Create a diagram of a computational graph.

    This is extremely useful for understanding and debugging computational
    graphs.  The structure and current state is converted to a PyGraphviz
    graph.  By default, each task is rendered as a box with the task label.
    Arrows are drawn between tasks, pointing from upstream producers of data to
    downstream consumers.  Arrows are labelled to show named inputs, if any.
    The color of each box shows its state: white for unfinished tasks, red for
    tasks that are failed, and black for tasks that are finished.

    Callers can customize the appearance of the graph by modifying the result
    before rendering it to an image or Jupyter notebook.

    Parameters
    ----------
    graph: :class:`graphcat.graph.Graph` derivative or :class:`pygraphviz.AGraph`, required
        The graph to be visualized.
    hide: Python callable, optional
        Python callable that can be used to hide tasks in the displayed figure.
        If :any:`None` (the default), all tasks will be displayed.  Ignored if
        `graph` is an instance of :class:`pygraphviz.AGraph`.
    rankdir: :class:`str`, optional
        Graphviz rankdir attribute that determines the direction of data flow
        within the diagram.  Default: ``"LR"``, which is left-to-right flow.
        Ignored if `graph` is an instance of :class:`pygraphviz.AGraph`.

    Returns
    -------
    diagram: :class:`pygraphviz.AGraph`
        Diagrammatic representation of `graph`.  Callers can modify `diagram`
        as needed before using its layout and drawing methods to produce a
        final image.

    See Also
    --------
    :func:`graphcat.notebook.display` - displays a graph in a Jupyter notebook.
    """

    if isinstance(graph, pygraphviz.AGraph):
        return graph

    if hide is None:
        hide = none

    nodes = [node for node in graph._graph.nodes() if not hide(graph, node)]
    subgraph = graph._graph.subgraph(nodes)

    black = "#494744"
    red = "crimson"
    white = "white"

    edgestyle = "solid"
    arrowhead = "lnormal" if graph.is_streaming else "normal"
    arrowhead = "o" + arrowhead if graph.is_dynamic else arrowhead

    agraph = pygraphviz.AGraph(directed=True, strict=False, ranksep="0.4", rankdir=rankdir)
    agraph.node_attr.update(fontname="Helvetica", fontsize=8, shape="box", style="filled", margin="0.08,0.04", width="0.4", height="0")
    agraph.edge_attr.update(fontname="Helvetica", fontsize=8, color=black, arrowhead=arrowhead, style=edgestyle)

    for node in subgraph.nodes():
        if subgraph.nodes[node]["state"] == graphcat.TaskState.UNFINISHED:
            color = black
            fontcolor = black
            fillcolor = white
        if subgraph.nodes[node]["state"] == graphcat.TaskState.FAILED:
            color = red
            fontcolor = white
            fillcolor = red
        if subgraph.nodes[node]["state"] == graphcat.TaskState.FINISHED:
            color = black
            fontcolor = white
            fillcolor = black

        agraph.add_node(node, color=color, fillcolor=fillcolor, fontcolor=fontcolor)

    for target, source, input in subgraph.edges(data="input"):
        if input is None:
            input = ""
        agraph.add_edge(source, target, label=f"  {input}  ") # We want edges to point from dependencies to dependents.

    return agraph


def leaves(graph, node):
    """Filter function that hides all leaf nodes when displaying a graph using :func:`draw`."""
    return graph._graph.out_degree(node) == 0


def none(graph, node):
    """Do-nothing filter function used to display an entire graph using :func:`draw`."""
    return False


def parameters(graph, node):
    """Filter function that hides "parameter" nodes."""
    return isinstance(graph._graph.nodes[node]["fn"], (graphcat.Array, graphcat.Constant))


def performance(agraph, monitor):
    """Add performance monitor information to a graph diagram.

    Parameters
    ----------
    agraph: :class:`pygraphviz.AGraph`, required
        Diagram originally created using :func:`draw`.
    monitor: :class:`graphcat.common.PerformanceMonitor`, required
        Performance monitor object containing performance results to be
        added to `agraph`

    Returns
    -------
    diagram: :class:`pygraphviz.AGraph`
        Input diagram supplemented with performance results from `monitor`.
    """
    all_times = [times[-1] for times in monitor.tasks.values()]
    min_time = min(all_times)
    max_time = max(all_times)

    agraph = agraph.copy()
    agraph.graph_attr["forcelabels"] = True
    for name, times in monitor.tasks.items():
        time = times[-1]
        if max_time - min_time > 0:
            percent = (time - min_time) / (max_time - min_time)
            if percent > 0.66:
                timecolor = "red"
            elif percent > 0.33:
                timecolor = "#ffaa00"
            else:
                timecolor = "green"
        else:
            timecolor="black"
        agraph.get_node(name).attr["xlabel"] = f"<<font color='{timecolor}'>&#11044;</font> <font color='black'>{time:.4f}s</font>>"
    return agraph





# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Abstract interfaces for computational graphs.
"""

import abc

import blinker
import networkx

import graphcat.common

class Graph(abc.ABC):
    """Abstract base class for computational graphs.

    The graph is a collection of named tasks, connected by links that define
    dependencies between tasks.  Updating a task implicitly updates all of its
    transitive dependencies.  When an unfinished task is updated, it executes a
    user-supplied function and stores the function return value as the task
    output.  Outputs of upstream tasks are automatically passed as inputs to
    downstream tasks.
    """
    def __init__(self):
        self._graph = networkx.MultiDiGraph()
        self._on_changed = blinker.Signal()
        self._on_cycle = blinker.Signal()
        self._on_execute = blinker.Signal()
        self._on_failed = blinker.Signal()
        self._on_finished = blinker.Signal()
        self._on_task_renamed = blinker.Signal()
        self._on_update = blinker.Signal()


    @abc.abstractmethod
    def _add_node(self, name, fn):
        raise NotImplementedError() # pragma: no cover


    def __contains__(self, name):
        return name in self._graph


    @abc.abstractmethod
    def _mark_unfinished(self, name):
        raise NotImplementedError() # pragma: no cover


    def _require_valid_names(self, names):
        if names is None:
            return self.tasks()
        if not isinstance(names, (list, set)):
            names = [names]
        return {name for name in names if name in self._graph}


    def _require_task_present(self, name):
        if name not in self._graph:
            raise ValueError(f"Task {name!r} doesn't exist.")


    def _require_task_absent(self, name):
        if name in self._graph:
            raise ValueError(f"Task {name!r} already exists.")


    def add_links(self, source, targets):
        """Add links between `source` and `targets`.

        Note that calling :meth:`add_links` multiple times will create
        multiple, parallel links between tasks.

        See Also
        --------
        :meth:`set_links`, which is the idempotent alternative to :meth:`add_links`.

        Parameters
        ----------
        source: hashable object, required
            Name of the task that will act as a data source.
        targets: tuple, or list of tuples, required
            Each (task, input) tuple specifies the target of a link.

        Raises
        ------
        :class:`ValueError`
            If `source` or `target` don't exist.
        """
        self._require_task_present(source)

        if not isinstance(targets, list):
            targets = [targets]

        # Add new edges
        unfinished = set()
        for target in targets:
            if isinstance(target, tuple):
                target, input = target
            else:
                input = None
            self._require_task_present(target)
            self._graph.add_edge(target, source, input=input) # Edges point from tasks to their dependencies.
            unfinished.add(target)

        self.mark_unfinished(unfinished)


    def add_task(self, name, fn=None):
        """Add a task to the graph.

        This function will raise an exception if the task already exists.

        See Also
        --------
        :meth:`set_task` : idempotent alternative to :meth:`add_task`.

        Parameters
        ----------
        name: hashable object, required
            Unique label that will identify the task.
        fn: callable, optional
            The `fn` object will be called whenever the task is executed.  It must take two keyword arguments
            as parameters, `label` and `inputs`.  `name` will contain the unique task name.  `inputs` will
            be a dict mapping named inputs to a sequence of outputs returned from upstream tasks.
            If :any:`None` (the default), :func:`graphcat.common.null` will be used.

        Raises
        ------
        :class:`ValueError`
            If `label` already exists.
        """
        self._require_task_absent(name)
        if fn is None:
            fn = graphcat.common.null
        self.set_task(name, fn)


    def clear_links(self, source, target):
        """Remove links from the graph.

        This method will remove all links from `source` to `target`.

        Parameters
        ----------
        source: hashable object, required
            Source task name.
        target: hashable object, required
            Target task name.

        Raises
        ------
        :class:`ValueError`
            If `source` or `task` don't exist.
        """
        self._require_task_present(source)
        self._require_task_present(target)
        self.mark_unfinished(source)
        while self._graph.number_of_edges(target, source):
            self._graph.remove_edge(target, source)


    def clear_tasks(self, names=None):
        """Remove tasks from the graph, along with all related links.

        Note that downstream tasks will become unfinished.

        Parameters
        ----------
        names: :any:`None`, hashable object, or list|set of hashable objects, optional
            Names identifying the tasks to remove.  If :any:`None` (the default), all tasks are removed, emptying the graph.
        """
        names = self._require_valid_names(names)
        self.mark_unfinished(names)
        for name in names:
            self._graph.remove_node(name)
        self._on_changed.send(self)


    @property
    @abc.abstractmethod
    def is_dynamic(self):
        """Return :any:`True` if-and-only-if the graph implements dynamic updates."""
        raise NotImplementedError() # pragma: no cover


    @property
    @abc.abstractmethod
    def is_streaming(self):
        """Return :any:`True` if-and-only-if the graph implements streaming (extent-based) updates."""
        raise NotImplementedError() # pragma: no cover


    def links(self, names=None):
        """Return every link originating with the given names.

        Parameters
        ----------
        names: :any:`None`, hashable object, or list|set of hashable objects, optional
            Names identifying the source tasks from which to retrieve links.

        Returns
        -------
        links: :class:`list`
            (source, target, input) tuple for every link in the graph.
        """
        names = self._require_valid_names(names)

        results = []
        for target, source, input in self._graph.edges(data="input"):
            if source in names:
                results.append((source, (target, input)))
        return results


    def mark_unfinished(self, names=None):
        """Set the unfinished state for tasks and all downstream dependents.

        Normally, the unfinished state is set automatically when changes are
        made to the graph.  This method is provided for callers who need to
        set the unfinished state in response to some outside event that the
        graph isn't aware of; this should only happen in extremely rare
        situations.

        Parameters
        ----------
        names: :any:`None`, hashable object, or list|set of hashable objects, required
            Task names to be marked as unfinished.  If :any:`None` (the default), the entire graph is marked unfinished.
        """
        names = self._require_valid_names(names)

        for name in list(names):
            for ancestor in networkx.ancestors(self._graph, name):
                names.add(ancestor)

        for name in names:
            self._mark_unfinished(name)

        self._on_changed.send(self)


    @property
    def on_changed(self):
        """Signal emitted whenever a part of the graph becomes unfinished.

        Functions invoked by this signal must have the signature fn(graph),
        where `graph` is this object.

        Returns
        -------
        signal: :class:`blinker.base.Signal`
        """
        return self._on_changed


    @property
    def on_cycle(self):
        """Signal emitted if a cycle is detected during updating.

        Functions invoked by this signal must have the signature fn(graph, name),
        where `graph` is this object.

        Returns
        -------
        signal: :class:`blinker.base.Signal`
        """
        return self._on_cycle


    @property
    def on_execute(self):
        """Signal emitted before a task is executed.

        Functions invoked by this signal must have the signature fn(graph, name, input),
        where `graph` is this object, `name` is the name of the task to be
        executed, and `input` is a dict containing the task inputs.

        Returns
        -------
        signal: :class:`blinker.base.Signal`
        """
        return self._on_execute


    @property
    def on_failed(self):
        """Signal emitted when a task fails during execution.

        Functions invoked by this signal must have the signature fn(graph, name, exception),
        where `graph` is this object, `name` is the name of the task that failed, and
        `exception` is the exception raised by the task.

        Returns
        -------
        signal: :class:`blinker.base.Signal`
        """
        return self._on_failed


    @property
    def on_finished(self):
        """Signal emitted when a task executes successfully.

        Functions invoked by this signal must have the signature fn(graph, name, output),
        where `graph` is this object, `name` is the name of the task that
        executed successfully, and `output` is the return value from the task
        function.

        Returns
        -------
        signal: :class:`blinker.base.Signal`
        """
        return self._on_finished


    @property
    def on_task_renamed(self):
        """Signal emitted when a task is renamed.

        Functions invoked by this signal must have the signature
        fn(graph, oldname, newname), where `graph` is this object, `oldname` is
        the original name of the task, and `newname` is its
        current name.

        Returns
        -------
        signal: :class:`blinker.base.Signal`
        """
        return self._on_task_renamed


    @property
    def on_update(self):
        """Signal emitted when a task is updated.

        Functions invoked by this signal must have the signature fn(graph, name),
        where `graph` is this object and `name` is the name of the task to be
        updated.

        Returns
        -------
        signal: :class:`blinker.base.Signal`
        """
        return self._on_update


    @abc.abstractmethod
    def output(self, name):
        """Retrieve the output from a task.

        This implicitly updates the graph, so the returned value is
        guaranteed to be up-to-date.

        Parameters
        ----------
        name: hashable object, required
            Unique task name.

        Returns
        -------
        output: any object
            The value returned when the task function was last executed, or :any:`None`.

        Raises
        ------
        :class:`ValueError`
            If `name` doesn't exist.
        :class:`Exception`
            Any exception raised by a task function will be re-raised by :meth:`output`.
        """
        raise NotImplementedError() # pragma: no cover


    def rename_task(self, oldname, newname):
        """Change an existing task's name.

        This modifies an existing task's name and modifies any related links
        as-necessary.  In addition, the task and any downstream dependents will
        become unfinished.

        Parameters
        ----------
        oldname: hashable object, required
            Existing original task name.
        newname: hashable object, required
            Unique new task name.

        Raises
        ------
        :class:`ValueError`
            If the task with `oldname` doesn't exist, or a task with `newname` already exists.
        """
        self._require_task_present(oldname)
        self._require_task_absent(newname)
        networkx.relabel_nodes(self._graph, mapping = {oldname: newname}, copy=False)
        self.mark_unfinished(newname)
        self._on_task_renamed.send(self, oldname=oldname, newname=newname)


    def set_expression(self, name, expression, symbols=None):
        """Create a task that evaluates a Python expression, returning its value.

        The task will automatically track implicit dependencies that
        arise from executing the expression.

        Parameters
        ----------
        name: hashable object, required
            Unique name for the new expression task.
        expression: string, required
            Python expression that will be evaluated whenever the task is executed.
        symbols: callable, optional
            Function that returns a Python dict containing symbols that will be
            available to the expression when it's executed.  If :any:`None` (the
            default), the :func:`graphcat.common.builtins` function will be used, which gives the
            expression access to `graph`, `name`, `inputs`, and `extent` objects
            that match the arguments to a normal task function.
        """
        fn = graphcat.common.evaluate(expression, symbols)
        fn = graphcat.common.automatic_dependencies(fn)
        self.set_task(name, fn)


    def set_links(self, source, targets):
        """Set links between `source` and `targets`.

        .. note::

            This function overrides *all* links from `source`.

        Parameters
        ----------
        source: hashable object, required
            Name of the task that will act as a data source.
        targets: tuple, or list of tuples, required
            Each (task, input) tuple specifies the target of a link.

        Raises
        ------
        :class:`ValueError`
            If `source` or `target` don't exist.
        """
        self._require_task_present(source)

        if not isinstance(targets, list):
            targets = [targets]

        # Remove existing edges
        unfinished = set()
        for remove_target, remove_source in list(self._graph.in_edges(source)):
            self._graph.remove_edge(remove_target, remove_source)
            unfinished.add(remove_target)

        # Add new edges
        for target in targets:
            if isinstance(target, tuple):
                target, input = target
            else:
                input = None
            self._require_task_present(target)
            self._graph.add_edge(target, source, input=input) # Edges point from tasks to their dependencies.
            unfinished.add(target)

        self.mark_unfinished(unfinished)


    def set_parameter(self, target, input, source, value):
        """Create and link a 'parameter' task in one step.

        Because they're so ubiquitous, this method simplifies the creation of
        "parameter" tasks - tasks that return a value for use as a parameter in
        some other task.  It consolidates creating the parameter task and
        linking it with an existing computational task into one step.

        Parameters
        ----------
        target: hashable object, required
            Name of the task that will use the parameter.
        input: hashable object, required
            Named input that will receive the parameter.
        source: hashable object, required
            Name of the task that will store the parameter.
        value: any Python object, required
            Parameter value.

        Raises
        ------
        :class:`ValueError`
            If `target` doesn't exist.
        """
        self.set_task(source, graphcat.common.constant(value))
        self.set_links(source, (target, input))


    def set_task(self, name, fn):
        """Add a task to the graph if it doesn't exist, and set its task function.

        Note that this will mark downstream tasks as unfinished.

        Parameters
        ----------
        name: hashable object, required
            Unique name that will identify the task.
        fn: callable, required
            The `fn` object will be called whenever the task is executed.  It must take two keyword arguments
            as parameters, `name` and `inputs`.  `name` will contain the unique task name.  `inputs` will
            be a dict mapping named inputs to sequences of outputs returned from upstream tasks.
        """
        if name in self._graph:
            if self._graph.nodes[name]["fn"] != fn:
                self.mark_unfinished(name)
            self._graph.nodes[name]["fn"] = fn
        else:
            self._add_node(name, fn=fn)
            self.mark_unfinished(name)


    def state(self, name):
        """Return the current state of a task.

        Parameters
        ----------
        name: hashable object, required
            Unique name that identifies the task.

        Returns
        -------
        state: :class:`graphcat.common.TaskState`
            Enumeration describing the current task state.

        Raises
        ------
        :class:`ValueError`
            If `name` doesn't exist.
        """
        self._require_task_present(name)
        return self._graph.nodes[name]["state"]


    def tasks(self):
        """Return the name of every task in the graph.

        Returns
        -------
        tasks: :class:`set`
            Names for every task in the graph.
        """
        return set(self._graph.nodes)


    @abc.abstractmethod
    def update(self, name):
        """Update a task and all its transitive dependencies.

        Parameters
        ----------
        name: hashable object, required
            Name identifying the task to be updated.

        Raises
        ------
        :class:`ValueError`
            If the task with `name` doesn't exist.
        :class:`Exception`
            Any exception raised by a task function will be re-raised by :meth:`update`.
        """
        raise NotImplementedError() # pragma: no cover



# Copyright 2020 Timothy M. Shead
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Integration with Jupyter notebooks, https://jupyter.org"""

import graphcat.diagram
import graphcat.optional
import graphcat.require


IPython = graphcat.optional.module("IPython.display")


@graphcat.require.loaded_module("IPython.display")
def display(graph, hide=None, rankdir="LR"):
    """Display a computational graph inline in a Jupyter notebook.

    This is extremely useful for understanding and debugging graphs.  The
    structure and current state of the graph is displayed as an inline SVG
    graphic.  See :func:`graphcat.diagram.draw` for details.

    Parameters
    ----------
    graph: :class:`graphcat.Graph` or :class:`pygraphviz.AGraph`, required
        The graph to be visualized.
    hide: Python callable, optional
        Python callable that can be used to hide tasks in the displayed figure.
        If :any:`None` (the default), all tasks will be displayed.  Ignored if
        `graph` is a :class:`pygraphviz.AGraph`.
    rankdir: :class:`str`, optional
        Graphviz rankdir attribute that determines the direction of data flow
        within the diagram.  Default: ``"LR"``, which is left-to-right flow.
        Ignored if `graph` is a :class:`pygraphviz.AGraph`.
    """

    agraph = graphcat.diagram.draw(graph, hide=hide, rankdir=rankdir)
    IPython.display.display(IPython.display.SVG(data=agraph.draw(prog="dot", format="svg")))
